import logging
import os
import traceback
from dataclasses import dataclass, field
from datetime import datetime, timedelta
from inspect import getframeinfo, stack
from pathlib import Path
from typing import Optional, override

from core.Utils import percentage
from core.domain.map.RouteInfo import RouteInfo
from core.domain.optimization.Optimization import Optimization
from core.domain.optimization.OptimizationFiles import OptimizationFiles
from core.domain.optimization.OptimizationMetrics import OptimizationMetrics
from core.domain.optimization.OptimizationPoint import OptimizationPoint
from core.domain.optimization.OptimizationPointType import OptimizationPointType
from core.domain.optimization.OptimizationResult import OptimizationResult
from core.domain.optimization.OptimizationResultData import OptimizationResultData
from core.domain.optimization.OptimizationRoute import OptimizationRoute
from core.domain.optimization.OptimizationSolution import OptimizationSolution
from core.domain.optimization.OptimizationState import OptimizationState
from core.domain.optimization.TransportMode import TransportMode
from core.domain.worker.WorkerJob import WorkerJob
from core.domain.worker.WorkerLogLevel import WorkerLogLevel
from core.repos.OptimizationMetricsRepo import OptimizationMetricsRepo
from core.repos.OptimizationRepo import OptimizationRepo
from core.repos.OptimizationResultRepo import OptimizationResultRepo
from core.repos.OptimizationVehicleRepo import OptimizationVehicleRepo
from core.repos.PostOfficeRepo import PostOfficeRepo
from core.repos.WorkerJobLogRepo import WorkerJobLogRepo
from core.repos.WorkerLogRepo import WorkerLogRepo
from core.types.Logger import Logger
from core.services.OptimizationService import OptimizationService
from core.services.RoutingService import RoutingService
from core.services.SystemService import SystemService
from core.types.Id import Id
from core.usecases.logging.Log_worker_job_status import Log_worker_job_status
from core.usecases.optimization_worker.Read_optimization_files import Read_optimization_files
from core.usecases.optimization_worker.Update_optimization_points import Update_optimization_points

logger = logging.getLogger(__name__)


@dataclass
class Run_optimization_job:
	workerLogRepo: WorkerLogRepo
	workerJobLogRepo: WorkerJobLogRepo
	log_worker_job_status: Log_worker_job_status
	read_optimization_files: Read_optimization_files
	update_optimization_points: Update_optimization_points
	optimizationVehicleRepo: OptimizationVehicleRepo
	optimizationRepo: OptimizationRepo
	postOfficeRepo: PostOfficeRepo
	optimizationService: OptimizationService
	optimizationResultRepo: OptimizationResultRepo
	routingService: RoutingService
	systemService: SystemService
	optimizationMetricsRepo: OptimizationMetricsRepo

	@dataclass
	class TerminationCallback:
		ctx: 'Run_optimization_job'
		optimization: Optimization
		pid: int
		log: Logger

		def __call__(self):
			optimizationState = self.ctx.optimizationRepo.get(id=self.optimization.id).state
			terminate = optimizationState != OptimizationState.ACTIVE
			if terminate:
				self.log.info(f"Optimization terminated: PID={self.pid}")
				self.ctx.systemService.killProcess(pid=self.pid)
			return False

	@dataclass
	class WorkerLogger(Logger):
		ctx: 'Run_optimization_job'
		workerJob: WorkerJob
		workerLogRepo: WorkerLogRepo
		workerJobLogRepo: WorkerJobLogRepo

		def log(self, level: int, workerLogLevel: WorkerLogLevel, data):
			dataStr = str(data)
			caller = getframeinfo(stack()[2][0])
			fileName = Path(caller.filename)
			context = f"{fileName.name}:{caller.lineno}"
			logger.log(level, dataStr, stacklevel=3)
			self.workerLogRepo.post(context=context, workerId=self.workerJob.workerId, data=dataStr, level=workerLogLevel)
			self.workerJobLogRepo.post(context=context, workerJobId=self.workerJob.id, data=dataStr, level=workerLogLevel)

		@override
		def info(self, data):
			self.log(logging.INFO, WorkerLogLevel.INFO, data)

		@override
		def warning(self, data):
			self.log(logging.WARNING, WorkerLogLevel.WARN, data)

		@override
		def error(self, data):
			self.log(logging.ERROR, WorkerLogLevel.ERROR, data)

	@dataclass
	class SolutionCallback:

		ctx: 'Run_optimization_job'
		workerJob: WorkerJob
		optimization: Optimization
		optimizationFiles: OptimizationFiles
		log_worker_job_status: Log_worker_job_status

		log: Logger

		optimizationResultId: Id[OptimizationResult] = Id.field()
		lastSolutionProcessedAt: datetime = field(default_factory=datetime.now)
		lastObjective: Optional[int] = field(default=None)
		currentNumberOfDistricts: int = field(default=None)
		solutionCount: int = field(default=0)
		locked: bool = field(default=False)
		crn_optimizationPointIndex: dict[int, int] = field(default_factory=lambda: {})

		def __post_init__(self):
			self.log.info("Mapping optimization points")
			for i, op in enumerate(self.optimizationFiles.optimizationPoints):
				self.crn_optimizationPointIndex[op.crnPoint.hisa] = i

		def __call__(self, objective: int, optimizationSolution: list[OptimizationSolution], finished: bool, unvisited: list[OptimizationPoint],
		             overlapping: Optional[dict[int, float]]):
			try:
				self.solutionCount += 1
				if self.lastObjective is None:
					self.lastObjective = objective
				if self.currentNumberOfDistricts is None:
					self.currentNumberOfDistricts = len(optimizationSolution)

				diff = (datetime.now() - self.lastSolutionProcessedAt)

				# If number of district changed SAVE with PRIORITY!
				if self.currentNumberOfDistricts != len(optimizationSolution):
					self.currentNumberOfDistricts = len(optimizationSolution)

					# Save first optimization that has changed
					self.log.info(f"Solution has changed number of districts: {self.currentNumberOfDistricts}")
					self.log.info(f"Solution is accepted: {self.solutionCount}. age={diff}, cost={objective}, diff={objective - self.lastObjective}")

					self.optimizationResultId = Id()  # Make new optimization result (do not override it with the old one)
					self.save(objective=objective, optimizationSolution=optimizationSolution, finished=finished, unvisited=unvisited,
					          overlapping=overlapping)
					self.lastObjective = objective

					self.optimizationResultId = Id()  # Make new optimization result (do not override it with the old one)
					self.save(objective=objective, optimizationSolution=optimizationSolution, finished=finished, unvisited=unvisited,
					          overlapping=overlapping)
					self.lastObjective = objective

					self.log.info("Solution callback ready to accept new solutions")
					self.lastSolutionProcessedAt = datetime.now()
					return

				if self.locked:
					self.log.info("Callback is locked")
					return

				# Callback is not locked so lets lock it.
				self.locked = True

				# If optimization is not first nor final then terminate saving the solution if its too early to save it.
				if not finished and self.solutionCount > 1:
					# If solution age is less than minute or there is no solution improvement then ignore solution
					if diff < timedelta(minutes=1) or objective > self.lastObjective:
						# Execute return release the lock
						self.locked = False
						return

				self.log.info(f"Solution is accepted: {self.solutionCount}. age={diff}, cost={objective}, diff={objective - self.lastObjective}")
				self.save(objective=objective, optimizationSolution=optimizationSolution, finished=finished, unvisited=unvisited,
				          overlapping=overlapping)
				self.lastObjective = objective
				self.log.info("Solution callback ready to accept new solutions")
				self.lastSolutionProcessedAt = datetime.now()

			except Exception:
				self.log.error(traceback.format_exc())
				try:
					self.ctx.optimizationRepo.setState(id=self.optimization.id, toState=OptimizationState.FAILED)
				except:
					pass

			self.locked = False

		def save(self, objective: int, optimizationSolution: list[OptimizationSolution], finished: bool, unvisited: list[OptimizationPoint],
		         overlapping: Optional[dict[int, float]]):
			optimizationResult = OptimizationResult(
				id=self.optimizationResultId, optimizationId=self.optimization.id, routes=[],
				info=f"{'Finished' if finished else 'Unfinished'} optimization result v{self.solutionCount} d{len(optimizationSolution)}",
				authorizedByUserId=self.optimization.authorizedByUserId, parent=None,
				unvisited=unvisited)

			optimizationMetrics = OptimizationMetrics(
				optimizationId=self.optimization.id, solution=self.solutionCount,
				vehicles=len(optimizationSolution),
				cost=objective, duration=0, distance=0, createdAt=datetime.now(), overlapping=overlapping)

			for routeI, route in enumerate(optimizationSolution):
				# Inject depo as first point in the route manualy!!!
				route.hise.insert(0, route.hise[-1])

				vehicle = self.ctx.optimizationVehicleRepo.get(route.optimizationVehicleId)

				# Find legs of optimization result
				routePoints = [self.optimizationFiles.optimizationPoints[self.crn_optimizationPointIndex[hisa]] for hisa in route.hise]

				try:
					routeInfo = self.ctx.routingService.getRouteInfo(
						transportMode=vehicle.type,
						legs=[point.crnPoint.microLocation for point in routePoints]
					)
				except Exception:
					self.log.error(traceback.format_exc())
					return

				try:
					averageRouteInfo = self.ctx.routingService.getAverageRouteInfo(
						transportMode=vehicle.type,
						legs=[point.crnPoint.microLocation for point in routePoints],
						probability=[point.visitFrequency for point in routePoints],
						iterations=10
					)
				except Exception:
					self.log.error(traceback.format_exc())
					return

				serviceTimeFreqSum = sum([(rp.serviceTime.total_seconds() * rp.visitFrequency) for rp in routePoints])

				if finished:
					matrixRouteDuration = timedelta(seconds=0)
					matrixRouteDistance = 0
					for i in range(len(route.hise) - 1):
						startIndex = self.crn_optimizationPointIndex[route.hise[i]]
						endIndex = self.crn_optimizationPointIndex[route.hise[i + 1]]
						matrixRouteDuration += timedelta(
							seconds=float(self.optimizationFiles.routeMatrices[vehicle.type].durations[startIndex][endIndex]) / 1000
						)
						matrixRouteDistance += self.optimizationFiles.routeMatrices[vehicle.type].distances[startIndex][endIndex]

				# Transform
				optimizationRoute = OptimizationRoute(
					name=route.district if route.district is not None else f"O{routeI + 1}",
					isExtra=route.isExtra,
					index=routeI,
					cost=route.cost,

					# THIS MUST BE IN SYNC WITH Run_optimization_job.save WHERE OPTIMIZATION ROUTE IS CALCULATED!!!
					distance=routeInfo.distance,
					averageDistance=averageRouteInfo.distance,

					# THIS MUST BE IN SYNC WITH Run_optimization_job.save WHERE OPTIMIZATION ROUTE IS CALCULATED!!!
					duration=int(routeInfo.duration + serviceTimeFreqSum),
					averageDuration=int(averageRouteInfo.duration + serviceTimeFreqSum),

					vehicle=vehicle,
					points=routePoints,
					steps=routeInfo.steps,

				)

				optimizationRoute.setHash()
				optimizationResult.routes.append(optimizationRoute)

				optimizationMetrics.distance += optimizationRoute.distance
				optimizationMetrics.duration += optimizationRoute.duration

			# Sort optimization routes by name and fix index"
			optimizationResult.routes.sort(key=lambda r: r.name)
			for i, route in enumerate(optimizationResult.routes):
				route.index = i

			# Save optimization routes solution and metrics to database")
			self.ctx.optimizationResultRepo.post(optimizationResult=optimizationResult)
			self.ctx.optimizationMetricsRepo.post(optimizationMetrics=optimizationMetrics)
			self.log.info("Optimization solution successfully saved")

			if finished:
				self.log.info("Set optimization state to completed")
				self.ctx.workerLogRepo.post(context=__name__, workerId=self.workerJob.workerId,
				                            data=f"Saving optimization status: {self.optimization.posta}", level=WorkerLogLevel.INFO)
				self.ctx.optimizationRepo.setState(id=self.optimization.id, toState=OptimizationState.COMPLETED)

				self.log.info("End info logging")
				self.ctx.workerJobLogRepo.post(context=__name__, workerJobId=self.workerJob.id, data=f"Start worker job: {self.workerJob.id}",
				                               level=WorkerLogLevel.INFO)
				self.ctx.workerLogRepo.post(context=__name__, workerId=self.workerJob.workerId,
				                            data=f"Optimization finished: {self.optimization.posta}", level=WorkerLogLevel.INFO)

				self.log.info("Optimization successfully finished")

	def now(self, workerJob: WorkerJob):
		optimization: Optional[Optimization] = None
		log = self.WorkerLogger(ctx=self, workerJob=workerJob, workerLogRepo=self.workerLogRepo, workerJobLogRepo=self.workerJobLogRepo)

		try:
			log.warning(f"Start new optimization worker job: {workerJob}")

			log.info("Get all optimization informations")
			optimization: Optimization = self.optimizationRepo.get(id=workerJob.optimizationId)
			optimizationVehicles = self.optimizationVehicleRepo.getAllByOptimizationId(optimizationId=workerJob.optimizationId)
			log.info(f"Found optimization: {optimization}'")

			log.info("Get postOffice")
			postOffice = self.postOfficeRepo.get(posta=optimization.posta.value)
			log.info(f"Found post office: {postOffice}")

			log.info("Reading optimization files")
			optimizationFiles = self.read_optimization_files.now(posta=postOffice.posta)

			latestConfOptResultData = None
			if optimization.parent is not None:
				latestConfOptResult = self.optimizationResultRepo.getLatestByOptimizationId(optimizationId=optimization.parent)
				latestConfOptResultData = OptimizationResultData(optimization=optimization, optimizationResult=latestConfOptResult)
				log.info(f"Get latest confirmed optimization result: {latestConfOptResult.info}")
				log.info(f"Filter only CRN points from latest confirmed optimization result")
				for i, route in enumerate(latestConfOptResult.routes):
					routesCrnPoints = list(filter(lambda op: op.type == OptimizationPointType.CRN, route.points))
					log.info(f"{i}. Route crn points[{len(routesCrnPoints)}/{len(route.points)}]: {percentage(routesCrnPoints, route.points)}%")
					route.points = routesCrnPoints

			log.info("Update optimization files")
			optimizationFiles.optimizationPoints = self.update_optimization_points.now(
				optimization=optimization, optimizationPoints=optimizationFiles.optimizationPoints)

			log.info("Create solution callback")
			solutionCallback = self.SolutionCallback(ctx=self, workerJob=workerJob, optimization=optimization, optimizationFiles=optimizationFiles,
			                                         log_worker_job_status=self.log_worker_job_status, log=log)

			log.info("Create termination callback")
			terminationCallback = self.TerminationCallback(ctx=self, optimization=optimization, pid=os.getpid(), log=log)

			log.info("Start VRP optimization")
			self.optimizationService.vrpOptimization(
				optimization=optimization,
				optimizationVehicles=optimizationVehicles,
				optimizationPoints=optimizationFiles.optimizationPoints,
				routeMatrices=optimizationFiles.routeMatrices,
				solutionCallback=solutionCallback,
				terminationCallback=terminationCallback,
				log=log,
				initialOptimizationResultData=latestConfOptResultData,
			)

			log.info("Finish VRP optimization")

		except BaseException as err:
			if optimization is not None:
				self.optimizationRepo.setState(id=optimization.id, toState=OptimizationState.FAILED)
			log.error(traceback.format_exc())
			raise err
